//MPIparallelwrite.c
//Author: David Monismith
//
//Program to test writing data with parallel file IO

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define OUTPUT_STRING "3\n"
#define OUTPUT_BYTES 1024
#define STR_LEN 512
#define ONE_MB 1048576.0

int main(int argc, char ** argv)
{
  //Read in a file name from the command line
  char * file_name = argv[1];
  int num_ints = atoi(argv[2]); //Number of integers to write
  int rank;
  int npes;

  char * output_str = malloc(sizeof(char)*(OUTPUT_BYTES+1));

  //Declare MPI variables
  MPI_Info fileInfo = MPI_INFO_NULL;
  MPI_Offset fileOffset;
  MPI_File myFile;

  //Start MPI
  MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &npes);

    //Process rank 0 keeps track of time
    if(rank == 0)
    {
      MPI_Offset fSize = OUTPUT_BYTES*num_ints;
      MPI_File_open(MPI_COMM_WORLD, file_name, 
        MPI_MODE_CREATE | MPI_MODE_WRONLY | MPI_MODE_APPEND, fileInfo, &myFile);
      MPI_File_set_size(myFile, fSize);
      MPI_Barrier(MPI_COMM_WORLD);
      double startTime, endTime;
      startTime = MPI_Wtime();
      MPI_File_close(&myFile);
      MPI_Barrier(MPI_COMM_WORLD);
      endTime = MPI_Wtime();
      printf("The total runtime was %lf s.\n", (endTime - startTime));
      printf("The write speed was %lf MB/s.\n", ((num_ints*OUTPUT_BYTES)/ONE_MB)/(endTime - startTime));
    }
    else
    {
      //Open the file in all processes
      int i;
      int processNumInts = num_ints/(npes-1);
      if(rank == npes - 1)
        processNumInts += num_ints % (npes-1);

      output_str[0]='\0';
      for(i = 0; i < STR_LEN; i++)
        strcat(output_str, OUTPUT_STRING);

      MPI_File_open(MPI_COMM_WORLD, file_name, 
        MPI_MODE_CREATE | MPI_MODE_WRONLY | MPI_MODE_APPEND, fileInfo, &myFile);
      MPI_Barrier(MPI_COMM_WORLD);

      //Write data to the file using a blocking write

      for(i = 0; i < processNumInts; i++)
      {
        fileOffset = i*OUTPUT_BYTES + (rank-1)*OUTPUT_BYTES*processNumInts;
        MPI_File_write_at(myFile, fileOffset, output_str, OUTPUT_BYTES, MPI_CHAR, MPI_STATUS_IGNORE);
      }

      //Close the file

      MPI_File_close(&myFile);
      MPI_Barrier(MPI_COMM_WORLD);

    } 

  free((void *) output_str);
  MPI_Finalize();
}
